'Copyright (c) 2012-2013 Donald Milne
'All rights reserved.
'
'This file is part of ExtractIcons.
'
'ExtractIcons is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'ExtractIcons is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with ExtractIcons.  If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.Collections
Imports System.IO
Imports System.Runtime.CompilerServices
Imports System.Text

Public Class binFile

    Private fsStream As FileStream
    Private brReader As BinaryReader

    Public Sub New(ByVal FileName As String)
        fsStream = New FileStream(FileName, FileMode.Open, FileAccess.Read, FileShare.Read)
        brReader = New BinaryReader(fsStream)
    End Sub

    Public ReadOnly Property Stream() As BinaryReader
        Get
            Return brReader
        End Get
    End Property

End Class

Public Class RDBInterface

    Private prvDataFiles As New List(Of binFile)
    Private prvStrAOPath As String

    Public Sub New(ByRef strAOPath As String)
        Dim diDataDir As New DirectoryInfo(strAOPath & Path.DirectorySeparatorChar & "cd_image" & Path.DirectorySeparatorChar & "data" & Path.DirectorySeparatorChar & "db")
        Dim rxDBName As New System.Text.RegularExpressions.Regex("ResourceDatabase.dat")
        prvStrAOPath = strAOPath
        For Each fiFiles As FileInfo In diDataDir.GetFiles("ResourceDatabase.*")
            If rxDBName.IsMatch(fiFiles.FullName) Then prvDataFiles.Add(New binFile(fiFiles.FullName))
        Next
    End Sub

    Public Function DoExport(ByRef strDestPath As String, ByVal lngRecType As Int32) As Int32
        Return Me.Export(strDestPath, lngRecType)
    End Function

    Public Sub Dispose()
        For Each bfDataFile As binFile In prvDataFiles
            bfDataFile.Stream.Close()
        Next
    End Sub

    Private Function Export(ByRef strPath As String, ByVal lngRecType As Int32) As Int32
        Dim fsOutput As FileStream = New FileStream(strPath, FileMode.Create)
        Dim bwOutput As BinaryWriter = New BinaryWriter(fsOutput)
        'Dim fsDbgOutput As FileStream = New FileStream(strPath.Replace(".dat", ".txt"), FileMode.Create)
        'Dim swDbgOutput As StreamWriter = New StreamWriter(fsDbgOutput)
        Dim fsResIdx As FileStream = New FileStream(prvStrAOPath & Path.DirectorySeparatorChar & "cd_image" & Path.DirectorySeparatorChar & "data" & Path.DirectorySeparatorChar & "db" & Path.DirectorySeparatorChar & "ResourceDatabase.idx", FileMode.Open)
        Dim brResIdx As BinaryReader = New BinaryReader(fsResIdx)
        Dim intNumRecords As Int32 = 0
        Dim intForwardLink As UInt32 = 1
        Dim intBlockOffset As Int32 = 0
        Dim intFirstBlock As Int32 = 0
        Dim intDataFileSize As Int32 = 0
        Dim intActiveEntry As Int32 = 0
        Dim intROffset64 As UInt64 = 0
        Dim intROffset As UInt32 = 0
        Dim intROffset2 As UInt32 = 0
        Dim intRType As Int32 = 0
        Dim intCalculatedOffset As Int32 = 0
        Dim intDataSize As Int32 = 0
        Dim hdr(34) As Byte

        brResIdx.BaseStream.Seek(12L, SeekOrigin.Begin)
        intBlockOffset = brResIdx.ReadInt32()
        brResIdx.BaseStream.Seek(184L, SeekOrigin.Begin)
        intDataFileSize = brResIdx.ReadInt32()
        brResIdx.BaseStream.Seek(72L, SeekOrigin.Begin)
        intFirstBlock = brResIdx.ReadInt32()
        brResIdx.BaseStream.Seek(intFirstBlock, SeekOrigin.Begin)
        Do While intForwardLink > 0
            intForwardLink = brResIdx.ReadUInt32()
            ' skip BackLink
            brResIdx.ReadInt32()
            intActiveEntry = brResIdx.ReadInt16()
            ' skip 8 bytes of unknown
            brResIdx.ReadBytes(8)
            ' 1.1 : skip further 10 bytes of unknown
            brResIdx.ReadBytes(10)
            Do While intActiveEntry <> 0
                ' 1.1 : read additional 4 bytes as high Long
                intROffset2 = brResIdx.ReadUInt32()
                intROffset = brResIdx.ReadUInt32()
                ' 1.1 : recombine into a 64 bit offset
                intROffset64 = intROffset2 * (2 ^ 32) + intROffset
                intRType = Me.SwapEndian(brResIdx.ReadInt32())
                ' skip next
                brResIdx.ReadInt32()
                If (intRType = lngRecType) Then
                    hdr = ReadDataFile(34, intDataFileSize, intBlockOffset, intROffset64)
                    intDataSize = BitConverter.ToInt32(hdr, 18) - 12
                    If intDataSize > 0 And intDataSize < 10000 Then
                        bwOutput.Write(ReadDataFile(intDataSize, intDataFileSize, intBlockOffset, intROffset64 + 10))
                        'swDbgOutput.WriteLine("At Offset " & CStr(intROffset64 + 10) & " : " & intDataSize & " bytes")
                        'swDbgOutput.Flush()
                        intNumRecords = intNumRecords + 1
                    End If
                End If
                intActiveEntry = intActiveEntry - 1
            Loop
            brResIdx.BaseStream.Seek(intForwardLink, SeekOrigin.Begin)
        Loop
        brResIdx.Close()
        fsResIdx.Close()
        bwOutput.Close()
        'fsDbgOutput.Close()
        fsOutput.Close()
        Return intNumRecords
    End Function

    Private Function SwapEndian(ByVal lngIn As Int32) As Int32
        Dim bytes() As Byte = BitConverter.GetBytes(lngIn)
        If (BitConverter.IsLittleEndian) Then
            Array.Reverse(bytes)
        End If
        Return BitConverter.ToInt32(bytes, 0)
    End Function

    Public Sub SplitImages(ByRef strInputFile As String, ByRef strOutputPath As String)
        Dim fileStream1 As FileStream = New FileStream(strInputFile, FileMode.Open)
        Dim binaryReader As BinaryReader = New BinaryReader(fileStream1)

        Do While binaryReader.BaseStream.Position < binaryReader.BaseStream.Length
            Dim numArray1() As Byte = binaryReader.ReadBytes(4)
            Dim num1 As Int32 = binaryReader.ReadInt32()
            Dim num2 As Int32 = binaryReader.ReadInt32()
            Dim num3 As Int32 = binaryReader.BaseStream.Position
            Dim numArray2() As Byte = binaryReader.ReadBytes(12)
            Dim numArray3() As Byte = binaryReader.ReadBytes(6)
            If (binaryReader.ReadInt32() = 1179207242) Then
                binaryReader.BaseStream.Seek(-10L, SeekOrigin.Current)
                Dim fileStream2 As FileStream = New FileStream(strOutputPath & CStr(num1) + ".jfif", FileMode.Create)
                Dim binaryWriter As BinaryWriter = New BinaryWriter(fileStream2)
                binaryWriter.Write(binaryReader.ReadBytes(num2 - 12))
                fileStream2.Close()
                binaryWriter.Close()
            Else
                binaryReader.BaseStream.Seek(-10L, SeekOrigin.Current)
                Dim fileStream2 As FileStream = New FileStream(strOutputPath & CStr(num1) + ".png", FileMode.Create)
                Dim binaryWriter As BinaryWriter = New BinaryWriter(fileStream2)
                binaryWriter.Write(binaryReader.ReadBytes(num2 - 12))
                fileStream2.Close()
                binaryWriter.Close()
            End If
            binaryReader.BaseStream.Seek((num3 + num2 - 24), SeekOrigin.Begin)
        Loop
        fileStream1.Close()
        binaryReader.Close()
    End Sub

    Private Function ArrayMerge(ByVal Buffer1() As Byte, ByVal Buffer2() As Byte) As Byte()
        Dim Buffer3((Buffer1.Length + Buffer2.Length) - 1) As Byte
        Array.Copy(Buffer1, 0, Buffer3, 0, Buffer1.Length)
        Array.Copy(Buffer2, 0, Buffer3, Buffer1.Length, Buffer2.Length)
        Return Buffer3
    End Function

    'Private Function ReadDataFile(ByVal Count As Int32, ByVal DataFileSize As UInt32, ByVal BlockOffset As UInt32, Optional ByVal Offset As UInt32 = 0) As Byte()
    Private Function ReadDataFile(ByVal Count As Int32, ByVal DataFileSize As UInt32, ByVal BlockOffset As UInt32, Optional ByVal Offset As UInt64 = 0) As Byte()
        Static DataFileNum As UInt32
        If Offset <> 0 Then
            'calc which datafile the record is in
            DataFileNum = Math.Floor(Offset / DataFileSize)
            If DataFileNum > 0 Then Offset = (Offset - (DataFileSize * DataFileNum)) + BlockOffset
            prvDataFiles(DataFileNum).Stream.BaseStream.Seek(Offset, SeekOrigin.Begin)
        End If
        'calc how many bytes we can read from the file if not all
        Dim Size As UInt32 = prvDataFiles(DataFileNum).Stream.BaseStream.Length
        Dim Here As UInt32 = prvDataFiles(DataFileNum).Stream.BaseStream.Position
        If (Here + Count) <= Size Then Return prvDataFiles(DataFileNum).Stream.ReadBytes(Count)
        'else, we calc how many bytes to read then jump to the rest in the next DF
        Dim aHalf As UInt32 = Size - Here
        Dim Buffer1 As Byte() = prvDataFiles(DataFileNum).Stream.ReadBytes(aHalf)
        prvDataFiles(DataFileNum + 1).Stream.BaseStream.Seek(BlockOffset, SeekOrigin.Begin)
        Dim bHalf As UInt32 = Count - aHalf
        Dim Buffer2 As Byte() = prvDataFiles(DataFileNum + 1).Stream.ReadBytes(bHalf)
        Return ArrayMerge(Buffer1, Buffer2) 'Buffer3
    End Function

End Class
